package edu.harvard.mgh.purcell.gplink.base_form;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JTextField;

/**
 * An extention of JPanel which allows the selection of an alternate pheonotype
 * file.
 * @author Kathe Todd-Brown
 *
 */
@SuppressWarnings("serial")
class AlterPhenoPanel extends JPanel {

	/**
	 * Parent is the Forms that this panel is attached to.
	 */
	private Forms parent;
	/**
	 * A button that indicates the default phenotype defined in the
	 * input files is selected. 
	 */
	private GeneralUtil.RequiredSelection defalutButton;
	/**
	 * A button that indicates an alternate phenotype file should be used.
	 */
	private GeneralUtil.RequiredSelection alternateButton;
	/**
	 * A button that indicates an optional flag that selects the phenotype
	 * defined in a particular column of the alternative phenotypic file.
	 */
	private GeneralUtil.OptionalSelectButton mphenoButton;
	/**
	 * A text field that stores the file for the alternate phenotype(s).
	 */
	private JTextField phenoText;
	/**
	 * A text field that stores the integer for the column of the phenotypic
	 * we are interesterested in.
	 */
	private JTextField mpheno;
	/**
	 * A browser button that looks for .phe files and puts the file name in
	 * phenoText.
	 */
	private GeneralUtil.pickAFileButton browse;
	/**
	 * An ItemListener that enables/disenables the mpheno button based on
	 * if the alternative phenoe file is flaged.
	 */
	private ItemListener linkMphenoAltern = new ItemListener(){
		public void itemStateChanged(ItemEvent arg0) {
			mphenoButton.setEnabled(alternateButton.isSelected());
			if(alternateButton.isSelected() == false)
				mphenoButton.setSelected(false);
		}
		
	};
	
	/**
	 * This function generates command line for an alternate phenotype
	 * flag.
	 * @return A string with the alternate phenotype information
	 * flagged.
	 */
	protected String process(){
		String ans = "";
		
		if(defalutButton.isSelected())
			return ans;

		ans = " --pheno " 
			+ GeneralUtil.processFilename(phenoText.getText());
		
		if(mphenoButton.isSelected() == true 
				&& mpheno.getText().matches("\\d+")){
			ans += " --mpheno " + mpheno.getText();
		}
		return ans;
	}
	
	/**
	 * This fuction looks to see if the phenoText is of length non-zero
	 * if the default button is not selected.
	 * @return A boolean that tells weather or not the form is filled
	 * out with valid information.
	 */
	protected boolean validInput(){
		boolean ans = false;
		
		if(defalutButton.isSelected())
			ans = true;
		
		else 
			ans =  phenoText.getText().length() > 0;
	
		
		return ans;
	}
	
	/**
	 * Create the alternate phenotypic file command portion.
	 * @param p The Forms that this panel is part of.
	 */
	protected AlterPhenoPanel(Forms p){
		
		//********************
		//Initalize everything
		//********************
		super();
		parent = p;
		
		mpheno = new JTextField();
		mphenoButton = new GeneralUtil.OptionalSelectButton(
				"Filter for a column (--mpheno)", mpheno, null);
		mphenoButton.setEnabled(false);
		
		defalutButton = new GeneralUtil.RequiredSelection("Use default", null);
		defalutButton.setSelected(true);
		
		phenoText = new JTextField();
		phenoText.setEditable(false);

		browse = new GeneralUtil.pickAFileButton(".phe", 
				"Phenotypic Files", phenoText, parent, false);
		
		alternateButton = new GeneralUtil.RequiredSelection(
				"Use alternate phenotype: ", phenoText, browse, 
				linkMphenoAltern);
		
		//pick one!
		ButtonGroup groupie = new ButtonGroup();
		groupie.add(defalutButton);
		groupie.add(alternateButton);
		
		//******************
		//Lay everything out
		//******************
		setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.HORIZONTAL;
		
		c.gridwidth = 3;
		add(defalutButton, c);
		
		c.gridwidth = 1;
		c.gridy = 1;
		add(alternateButton, c);
		
		c.weightx = 1;
		c.gridx = 1;
		add(phenoText, c);
		
		c.weightx = 0;
		c.gridx = 2;
		add(browse, c);
		
		c.gridx = 0;
		c.gridy = 2;
		add(mphenoButton, c);
		
		c.gridx = 1;
		add(mpheno, c);
	}
	
}
