package edu.harvard.mgh.purcell.gplink.base_form;

import java.awt.Component;
import java.io.*;

//import javax.swing.JButton;
import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.TitledBorder;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import edu.harvard.mgh.purcell.gplink.data.*;

/**
 * An extention of JPanel that gets the output root from the user.
 * @author ktoddbrown
 */
@SuppressWarnings("serial")
public class OutFilePanel extends JPanel {
	
	private JTextField f;
	private Project data;
	
	private Forms parent;
	//public boolean validInput;
	
	/**
	 * Update the message that tells the user if the current
	 * output root is valid or not
	 */
	DocumentListener messageUpdate = new DocumentListener(){
		public void insertUpdate(DocumentEvent e) {
			feedback();
			parent.isValidCommand();
			//validInput = isValidInput();
		}
		public void removeUpdate(DocumentEvent e) {
			feedback();
			parent.isValidCommand();
			//validInput = isValidInput();
		}
		public void changedUpdate(DocumentEvent e) {
			feedback();
			parent.isValidCommand();
			//validInput = isValidInput();
		}
	};
	
	/**
	 * check to see if a string is all either a letter, number, "_", or "-"
	 * @param s string we are checking
	 * @return returns false if the string contains a 
	 * letter, number, "-" or "_"
	 */
	private boolean isAlphaNumeric(String s){
		//check that s is a valid string
		if(s == null || s.length() == 0)
			return false;
		
		//go through and make sure each characture in s is valid
		for(int i = 0; i < s.length(); i++){
			char c = s.charAt(i);
			if(!(Character.isLetter(c) 
					|| Character.isDigit(c)
					|| c == '-'
					|| c == '_'))
				return false;
		}
		
		return true;
	}
	
	public String getCommand(){
		String ans = f.getText();
		
		if(isValidInput() == false){
			JOptionPane.showMessageDialog(parent.myFrame,
				"Error trying to get output root from the current form. " +
				"You should never see this message.",
    			"gPLINK error",
    			JOptionPane.ERROR_MESSAGE);
			return null;
		}
		if(data.isRemote(false))
			return "--out " + data.getRemoteFile() + "/" +  GeneralUtil.processFilename(ans);
		else
			return "--out " + GeneralUtil.processFilename(ans);
	}
	
	/**
	 * check if a log file exists and is non-empty
	 * @return returns if the output log file exists and is non-empty
	 */
	public boolean isValidInput(){
		if(isAlphaNumeric(f.getText()) == false)
			return false;
		else{
			String[] fileList = data.getRoot().list();
			for(int i = 0; i < fileList.length; i ++){
				if(fileList[i].split("\\.")[0].equals(f.getText()))
					return false;
			}
			
		}
		
		return true;
		
	}
	
	/**
	 * set message to reflect the validity of a output root file
	 */
	private void feedback(){
		if(isValidInput())
			setBorder(new TitledBorder("Output file root: Valid fileroot"));
		else
			setBorder(new TitledBorder("Output file root: Invalid fileroot"));
		
	}
	
	/**
	 * Create a panel to get the output file data from the user
	 * @param p the project this panel is tied to
	 */
	public OutFilePanel(Project p, Forms givenF){
		
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		setBorder(new TitledBorder("Output file root: Invalid fileroot"));
		
		data = p;
		parent = givenF;
		
		
		//this is the text field that user will input the output root to
		f = new JTextField(40);
		//we want the user to be able to input the root
		f.setEditable(true);
		//this let's us give feed back on the chosen root
		f.getDocument().addDocumentListener(messageUpdate);
		
		
		JPanel filePanel = new JPanel();
		filePanel.setLayout(new BoxLayout(filePanel, BoxLayout.Y_AXIS));
		filePanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		JLabel rootname = new JLabel(data.getRoot().getAbsolutePath() 
				+ File.separator);
		filePanel.add(rootname);
		filePanel.add(f);
		this.add(filePanel);
		
	}
}
