package edu.harvard.mgh.purcell.gplink.data;

import org.w3c.dom.*;
import org.xml.sax.*;

import ch.ethz.ssh2.Connection;
import ch.ethz.ssh2.SCPClient;
import ch.ethz.ssh2.Session;
import ch.ethz.ssh2.StreamGobbler;

import javax.swing.JOptionPane;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import java.io.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

import edu.harvard.mgh.purcell.gplink.mainGUI.*;
import edu.harvard.mgh.purcell.gplink.base_form.GeneralUtil;
/**
 * This contains the data for using our operation traking.
 * @author Kathe Todd-Brown
 *
 */
public class Project {
	
	///////////////////////////////////////////////////////
	//////////////////// Static variables /////////////////
	///////////////////////////////////////////////////////

	/* ************************
	 * Default values
	 * ************************
	 */
	
	/** 
	 * default editor command for non-window & non-mac os
	 * */
	private static String LINUX_DEFAULT_EDITOR = "emacs";
	
	/** 
	 * default editor command for window os 
	 */
	private static String WINDOW_DEFAULT_EDITOR = "write";
	
	/** 
	 * default editor command for mac os 
	 */
	private static String MAC_DEFAULT_EDITOR = "open -a /Applications/TextEdit.app";
	
	/** 
	 * base of locked code in lock file that appear in LOCK_FILE
	 */
	private static String LOCKED_BASE = "True";
	
	/** 
	 * a unique lock code for this instance of 
	 * Project that appear in LOCK_FILE
	 */
	private static String LOCKED_CODE = LOCKED_BASE 
			+ (int)(Math.random()*100000000);
	
	/** 
	 * unlocked code that appear in LOCK_FILE
	 */
	private static String UNLOCKED_CODE = "False";
	
	/**
	 * mode string used when scp is called
	 */
	public String SCP_MODE = "0644";
	
	/* *****************************
	 * names of configationfiles
	 * *****************************
	 */
	
	/** 
	 * name of the meta file 
	 */
	public static String LOG_FILE = ".metafile_gplink";
	
	/** 
	 * name of lock file 
	 */
	public static String LOCK_FILE = ".lock_gplink";
	
	/** 
	 * name of file that stores remote parameters. In this case
	 * it's .gplink_remote
	*/
	public static String REMOTE_FILE = ".gplink_remote";
	
	/**
	 * extention for the status of a given operation file. In
	 * this case it's .gplink
	 */
	public static String OP_STATUS_EXT = ".gplink";
	
	/**
	 * configuration file for the global settings. In this case
	 * it's .gplink_config
	 */
	public static String GLOBAL_CONFIG_FILE = ".gplink_config";
	
	/** 
	 * root name of the operations summary file 
	 * (appended with time stamp and .html) 
	 */
	public static String OP_SUM_FILE = "logfile-";
	
	/* *******************************************
	 * keys for config files, and project document
	 * *******************************************
	 */
	
	/** 
	 * key name of file key in project/metafile
	 */
	private static String FILE_KEY = "entity";
	
	/**
	 * key name of the operations in project 
	 */
	private static String OP_KEY = "operation";
	
	/** 
	 * key name of the command line attribute key of the operation 
	 */ 
	private static String CLINE_KEY = "cline";
	
	/**
	 * key name of the plink path attribute key in the project that
	 * holds the location of plink exicutable
	 */
	public static String PLINK_PATH_KEY = "plink_path";
	
	/**
	 * key for the haploview path
	 */
	public static String HAPLOVIEW_PATH_KEY = "haploview_path";
	
	/**
	 * key for the haploview appendment
	 */
	public static String HAPLOVIEW_APPEND = "haploview_append";
	
	/**
	 * key for the editor
	 */
	public static String USER_EDITOR_KEY = "editor";
	
	/**
	 * key for prefix of plink command user controlled
	 */
	public static String PLINK_PREFIX = "plink_prefix";
	
	/**
	 * name of the number of operations ever created in
	 * attribute key of project
	 */ 
	private static String EOP_KEY = "ever_ops";
	
	/**
	 * name of the xml format version being used in project
	 */
	private static String VS_KEY = "version";
	
	/**
	 * name of the element containing the input files for an operation
	 */
	public static String INFILE_KEY = "input";
	
	/**
	 * name of the number of files in input/output attribute key
	 */
	private static String NUM_FILE_KEY = "num_file";
	
	/**
	 * name of the element containing the output files for an operation
	 */
	public static String OUTFILE_KEY = "output";
	
	/**
	 * Key for the browse only flag.
	 */
	public static String BROWSE_KEY = "browse";
	
	/**
	 * key for the most recent project opened for the gobal config
	 * file
	 */
	public static String PROJECT1 = "project1";
	
	/**
	 * key for the 2nd most recent project opened for the gobal config
	 * file
	 */
	public static String PROJECT2 = "project2";
	
	/**
	 * key for the 3nd most recent project opened for the gobal config
	 * file
	 */
	public static String PROJECT3 = "project3";
	
	/**
	 * key for the 4nd most recent project opened for the gobal config
	 * file
	 */
	public static String PROJECT4 = "project4";
	
	/**
	 * key for the home directory
	 */
	protected static String HOME_KEY = "home";
	
	/**
	 * key for name attribute in project, operation and file
	 */
	protected static String NAME_KEY = "name";
	
	/**
	 * String to flag a false boolean
	 */
	public static String FALSE = "false";
	
	/**
	 * String to flag a true boolean
	 */
	public static String TRUE = "true";
	
	/* ***********************************
	 * keys for remote configureation file
	 * ***********************************
	 */
	/**
	 * key for port number
	 */
	public static String PORT_KEY = "port";
	
	/**
	 * key for the user's name in the remote config file
	 */
	public static String USER_KEY = "user";
	
	/**
	 * key for the host's name in the remote config file
	 */
	public static String HOST_KEY = "host";
	
	/**
	 * key for the remote project folder in the remote config file
	 */
	public static String REMOTE_PROJECT_KEY = "folder";
	
	///////////////////////////////////////////////////////
	///////////////Non-Static Variables ///////////////////
	///////////////////////////////////////////////////////

	/**
	 * contains all the file and operations information
	 */
	private Document projectInfo;
	
	/**
	 * contains the root folder this is all in
	 */
	private File rootfolder;
	
	//TODO collaps remotefolder/host/user/port into one Properties
	//...variable
	
	/**
	 * path to the project folder in the remote machine
	 */
	private String remotefolder;
	
	/**
	 * remote host
	 */
	private String host;
	
	/**
	 * remote user
	 */
	private String user;
	
	/**
	 * port for ssh connection
	 */
	private int port;

	/**
	 * is the session flagged as read only
	 */
	private boolean readonly;
	
	/**
	 * boolean flag connected remote sessions
	 */
	private boolean isConn;
	
	//TODO access conn only via getter/setter
	/**
	 * ssh connection
	 */
	Connection conn;
	
	/**
	 * frame that we are interacting with the project with
	 */
	public MainFrame frame;
	
	/**
	 * automatically checks for new files
	 */
	public AutoUpdater gplinkLog;
	
	///////////////////////////////////////////////////////
	///////////////  Setter Functions  ////////////////////
	///////////////////////////////////////////////////////

	/**
	 * Set the browse flag
	 * @param browse a boolean that states the if you want the session
	 * to be browse only (true) or fully accessed (false)
	 */
	public void setBrowse(boolean browse){
		readonly = browse;
	}
	
	/**
	 * Set the current host.
	 * @param h a string discribing the host name on a remote
	 * session
	 */
	void setHost(String h){
		host = h;
	}
	
	
	/**
	 * Set the current user.
	 * @param u a string discribing the user name on the remote
	 * session
	 */
	void setUser(String u){
		user = u;
	}
	
	/**
	 * Set the port number.
	 * @param p an int specifying the port number for a remote ssh
	 * connection
	 */
	void setPort(int p){
		port = p;
	}
	
	/**
	 * Set the Connection for the project.
	 * @param c the connection that will be used in the 
	 * remote connection
	 */
	public void setConnection(Connection c){
		conn = c;
		isConn = true;
	}

	/**
	 * Set the local project file without initalizing the
	 * entire project.
	 * @param given file that makes the local project file
	 */
	public void setLocalRoot(File given){
		rootfolder = given;
	}
	
	/**
	 * set the path of plink
	 * @param p absolute path name of the plink exicutable
	 */
	public void setPlinkPath(String p){
		projectInfo.getDocumentElement().setAttribute(PLINK_PATH_KEY, p);
	}
	
	/**
	 * Set the haploview path
	 * @param p absolute path name of the haploview .jar file
	 */
	public void setHaploviewPath(String p){
		projectInfo.getDocumentElement().setAttribute(HAPLOVIEW_PATH_KEY, p);
	}
	
	/**
	 * Set an append to the command call to haploview
	 * @param s a string to append to haploview call
	 */
	public void setHaploviewAppend(String s){
		projectInfo.getDocumentElement().setAttribute(HAPLOVIEW_APPEND, s);
	}
	
	/**
	 * Set a preference to the plink command call
	 * @param p a string to prefix to the plink call
	 */
	public void setPlinkPrefix(String p){
		projectInfo.getDocumentElement().setAttribute(PLINK_PREFIX, p);
	}
	
	/**
	 * Set the user defined editor. This is a command to open an editor
	 * @param s an editor command
	 */
	public void setEditor(String s){
		projectInfo.getDocumentElement().setAttribute(USER_EDITOR_KEY, s);
	}
	
	/**
	 * If a connection is open then reset the remote project file.
	 * @param given an absolute path to the remote project file
	 */
	public void setRemoteFile(String given){
	
		//check that the connection is there
		if(conn == null){
			new ErrorDialog(frame, "No connection is open therefor I can not reset the remote project file. \n" +
			"setRemoteFile@Project.java");
			return; 
		}
		
		//rest the remote project
		//setRemoteRoot(given);
		remotefolder = given;
		
		/*
		 * Save the remote project configuration to local project
		 */
		Properties config = new Properties();
		
		try{
			config.setProperty(REMOTE_PROJECT_KEY, remotefolder);
			config.setProperty(USER_KEY, getUser());
			config.setProperty(HOST_KEY, host);
			config.setProperty(PORT_KEY, new Integer(port).toString());
			config.store(new FileOutputStream(new File(getRoot(), REMOTE_FILE)), 
			"gPLINK remote configuration");
		} catch (IOException e) {
			new ErrorDialog(frame, "Error saving the remote configuration file. \n" +
			"setRemoteFile@Project.java");
		}
		
		if(	getBrowse() == false &&
				frame.data.avalibleLock() == false)/*{
			int choice = JOptionPane.showConfirmDialog(frame, 
					"Some other instance of gPLINK has the lock. " +
					"Would you like to steal the " +
					"lock?\n If you don't then you will be unable to add any new operations. ", 
					"Steal?", JOptionPane.YES_NO_OPTION);
			if(choice == JOptionPane.YES_OPTION){
				setBrowse(false);
				lockMeta();
			} else {
				setBrowse(true);
			}
		}else*/ if(getBrowse() == false){
			lockMeta();
		}

		frame.menuStuff.setMenuLock(true);
		
		
		//load the project in the new remote folder
		loadData();
		
		//copy log and gplink files from the host
		updateFiles();
		
		//if it's attached to a MainFrame
		if(frame != null){
			//and there's an active gplinkLog
			if(gplinkLog != null)
				//cancel the gplinkLog
				gplinkLog.cancel();
			//reset the gplinkLog
			gplinkLog = new AutoUpdater(this);
		} else {
			//otherwise set the log to null since there's
			//no need to update a null MainFrame
			gplinkLog = null;
		}
	}
	
	/**
	 * Close the remote connection and clear the remote project file parameter.
	 */
	public void closeRemote(){
		//if the connection is not null then close it
		if(conn != null)
			conn.close();
		isConn = false;
		//make sure the connection is null
		conn = null;
		//clear the remote project file parameter
		remotefolder = "";
		setBrowse(frame.menuStuff.browseOnly.isSelected());
		//setRemoteRoot("");
	}
	
	/**
	 * Lock the project file(s) so that no other instance of gPLINK will conflict.
	 * If the project is a remote project then lock both the local and remote files.
	 *
	 */
	public void lockMeta(){
		
		/*
		 * try to lock the local project file
		 */
		try {
			BufferedWriter out = new BufferedWriter(
					new FileWriter(new File(getRoot(), LOCK_FILE)));
			out.write(LOCKED_CODE);
			out.newLine();
			out.close();
			//activate the appropreate menu items
			if(frame!=null)
				frame.menuStuff.setMenuLock(true);
		} catch (IOException e) {
			new ErrorDialog(frame, "Error writing the lock file locally. \n" +
			"lockMeta@Project.java");
			//eprintStackTrace();
		}
		
		/*
		 *get the lock from the remote
		 */
		if( isRemote(true)){
			Session sess;
			try {
			sess = conn.openSession();
			sess.execCommand("echo \"" + LOCKED_CODE + "\" > " 
					+ GeneralUtil.makeSpacesHard(getRemoteFile()+"/" +LOCK_FILE));
			sess.close();
			} catch (IOException e) {
				new ErrorDialog(frame, "Error trying to write the lock file remotely.\n" +
				"lockMeta:Project.java");
				connectionLost();
			}
			/*SCPClient scopy = new SCPClient(conn);
			try {
				scopy.put(getRoot().getAbsolutePath() 
						+ File.separator + LOCK_FILE,
						GeneralUtil.makeSpacesHard(getRemoteFile()) );
			} catch (IOException e) {
				new ErrorDialog(frame, "Error trying to write the lock file remotely.\n" +
						"lockMeta:Project.java");
			}*/
		}
		
	}
	/**
	 * Unlock the project file(s) so that no other instance of gPLINK will confict.
	 * If the project is a remote project then unlock both the local and remote files.
	 *
	 */
	public void unlockMeta(){
		
		/*
		 * try to lock the local project file
		 */
		try {
			BufferedWriter out = new BufferedWriter(
					new FileWriter(new File(getRoot(), LOCK_FILE)));
			out.write(UNLOCKED_CODE);
			out.newLine();
			out.close();
			//deactivate the appropreate menu items
			if(frame != null)
				frame.menuStuff.setMenuLock(false);
		} catch (IOException e) {
			new ErrorDialog(frame, "Error writing the lock file locally. \n" +
			"unlockMeta@Project.java");
			//eprintStackTrace();
		}
		
		/*
		 *get the lock from the remote
		 */
		if( isRemote(true)){
			Session sess;
			try {
			sess = conn.openSession();
			sess.execCommand("echo \"" + UNLOCKED_CODE + "\" > " 
					+ GeneralUtil.makeSpacesHard(getRemoteFile()+"/" +LOCK_FILE));
			sess.close();
			} catch (IOException e) {
				new ErrorDialog(frame, "Error trying to write the lock file remotely.\n" +
				"unlockMeta:Project.java");
				connectionLost();
			}
			/*SCPClient scopy = new SCPClient(conn);
			try {
				scopy.put(getRoot().getAbsolutePath() 
						+ File.separator + LOCK_FILE,
						GeneralUtil.makeSpacesHard(getRemoteFile()) );
			} catch (IOException e) {
				new ErrorDialog(frame, "Error trying to write the lock file remotely.\n" +
						"lockMeta:Project.java");
			}
			*/
		}
	}
	
	/**
	 * Set the root file we are working in and load the new data
	 * file that we find there. Checks also to see that the given
	 * file is a directory. Note that this local root must be set
	 * before the remote project can be set.
	 * @param f this is the folder that we are working from
	 */
	public void setRoot(File f){
		
		if(f == null){
			rootfolder = null;
//			and there's an active gplinkLog
			if(gplinkLog != null)
				//cancel the gplinkLog
				gplinkLog.cancel();
			return;
		}
			
		
		//if the connection is open we want close it
		if(conn != null)
			closeRemote();
//		set the folder
		rootfolder = f;	
		//load the project
		loadData();
		
		
		//check that it is a valid folder
		if(f.isDirectory() == true){
			//check that it's not write protected
			if(f.canWrite() == true){
				//if it's attached to a MainFrame
				if(frame != null){
					//and there's an active gplinkLog
					if(gplinkLog != null)
						//cancel the gplinkLog
						gplinkLog.cancel();
					//reset the gplinkLog
					gplinkLog = new AutoUpdater(this);
				} else {
					//otherwise set the log to null since there's
					//no need to update a null MainFrame
					gplinkLog = null;
				}
				
				//if the lock is not avaliable
				if(getBrowse() == false
						&& avalibleLock() == false){
					//then give the user an option to steal it
					/*int choice = JOptionPane.showConfirmDialog(null, 
							"Some other instance of gPLINK has the lock. " +
							"Would you like to steal the " +
							"lock?\n If you don't then you will be unable to add any new operations. ", 
							"Steal?", JOptionPane.YES_NO_OPTION);
					//if they choose to take the lock
					if(choice == JOptionPane.YES_OPTION){
						setBrowse(false);
						//lock the project
						lockMeta();						
					} else {
						setBrowse(true);
					}*/
					if(frame != null)
						frame.menuStuff.setMenuLock(false);
				} else {
					//the lock was available so let's take it
					lockMeta();
					if(frame != null)
						frame.menuStuff.setMenuLock(true);
				}
				
				
			}/*else{
				new ErrorDialog(frame, 
						"You must have write promisions in this " +
						"folder to set it as a gPLINK project.\n Currently " +
						"this folder is write protected.");
			}*/
		} else
			new ErrorDialog(frame,"Trying to set the working project folder " +
					"to a non-directory");
	}
	
	
	
	///////////////////////////////////////////////////////
	/////////////// Getter Functions   ////////////////////
	///////////////////////////////////////////////////////
	
	public void connectionLost(){
		if(isConn == true){
			int choice = JOptionPane.showConfirmDialog(frame, 
					"There is a problem with the connection.\n" +
					"Would you like try to reconnect?",
					"Connection Error",
				    JOptionPane.YES_NO_OPTION);
			
			if(choice == JOptionPane.YES_OPTION){
				String local = getRoot().getAbsolutePath();
				String remote = getRemoteFile();
				String tempHost = host;
				String tempUser = getUser();
				int tempPort = port;
				closeRemote();
				setLocalRoot(new File(local));
				new ProjectDialog(frame, local, remote, tempHost, 
						tempUser, tempPort);
			}
			else{
				isConn=false;
			}
		}
	}
	
	/**
	 * A project is remote if it has a non-empty local project folder
	 * and has a config file for a remote project
	 */
	public boolean isRemote(boolean checkConn){
		if(getRoot() == null 
				|| new File(getRoot(), REMOTE_FILE).exists() == false)
			return false;
		else if(checkConn == false)
			return true;
		
		//check for connection
		if(isConn == true
				&& (conn != null
						&& conn.isAuthenticationComplete() == false)){
			int choice = JOptionPane.showConfirmDialog(frame, 
					"There is a problem with the connection.\n" +
					"Would you like to reconnect?",
					"Connection Error",
				    JOptionPane.YES_NO_OPTION);
			
			if(choice == JOptionPane.YES_OPTION){
				String local = getRoot().getAbsolutePath();
				String remote = getRemoteFile();
				String tempHost = host;
				String tempUser = getUser();
				int tempPort = port;
				closeRemote();
				setLocalRoot(new File(local));
				new ProjectDialog(frame, local, remote, tempHost, 
						tempUser, tempPort);
				return false;
			}
			else{
				isConn = false;
				return false;
			}
		}
		else
			return isConn;
		
		
	}
	
	/**
	 * Check to see if the lock is avalable for this project.
	 * @return true if the lock is avaliable
	 */
	public boolean avalibleLock(){
		
		if(getBrowse() == true)
			return false;
		
		/*
		 * get the lock from the remote
		 */
		if(isRemote(true)){
			
			try {
				Session sess = conn.openSession();
				sess.execCommand("less " + GeneralUtil.makeSpacesHard(getRemoteFile() 
						+ "/" + LOCK_FILE));
				InputStream stdout = new StreamGobbler(sess.getStdout());
				BufferedReader br = new BufferedReader(new InputStreamReader(stdout));
				String remoteLock = br.readLine();
				sess.close();
				if(remoteLock != null
						&& !remoteLock.equals(LOCKED_CODE) 
						&& remoteLock.contains(LOCKED_BASE))
					return false;
				
			} catch (IOException e){
				new ErrorDialog(frame, "Error trying to read the remote lock");
				connectionLost();
			}
				/*SCPClient scopy = new SCPClient(conn);
				scopy.get(GeneralUtil.makeSpacesHard(getRemoteFile() 
						+ "/" + LOCK_FILE), getRoot().getAbsolutePath());
			} catch (IOException e) {
				try {
					new File(getRoot(), LOCK_FILE).createNewFile();
					unlockMeta();
				} catch (IOException e1) {
					new ErrorDialog(frame, "Error ensuring local file exists. (1) \n " +
					"avaliableLock@Project.java");
				}
				//can't copy for some reason... 
				//assume file does not exist and is avaliable
			}
				 */
			
		}
		/*
		 * Ensure that the files exist
		 */
		try {
			//if the file doesn't exist
			new File(getRoot(), LOCK_FILE).createNewFile();
		} catch (IOException e) {
			new ErrorDialog(frame, "Error ensuring local file exists. (2) \n " +
					"avaliableLock@Project.java");
		}
		
		/*
		 * read in the lock setting
		 */
		String localLock = "";
		try{
			/*
			 * read the local lock
			 */
			BufferedReader in = new BufferedReader(
					new FileReader(new File(getRoot(), LOCK_FILE)));
			//the lock code is on the first line
			localLock = in.readLine();
			in.close();
		}catch(IOException e){
			new ErrorDialog(frame, "Error trying to read in local lock.\n" +
					"avalibleLock@Project.java");
		}
		
		//does the lock either unclaimed or claimed by the current project
		if (localLock != null
				&& !localLock.equals(LOCKED_CODE) 
				&& localLock.contains(LOCKED_BASE))
			return false;
		else
			return true;
		
		
	}
	
	/**
	 * Check to see if the project currently has the lock
	 * @return true if the lock key equals that of the current project
	 */
	public boolean hasMetaLock(){
		
		if(getBrowse() == true)
			return false;
		
		/*
		 * use less to look at the data in the file
		 */
		if( isRemote(true) ){
			try{
				Session sess = conn.openSession();
				sess.execCommand("less " + GeneralUtil.makeSpacesHard(getRemoteFile() 
						+ "/" + LOCK_FILE));
				InputStream stdout = new StreamGobbler(sess.getStdout());
				BufferedReader br = new BufferedReader(new InputStreamReader(stdout));
				String remoteLock = br.readLine();
				sess.close();
				if(remoteLock == null
						|| !remoteLock.equals(LOCKED_CODE))
					return false;
				
			} catch (IOException e){
				new ErrorDialog(frame, "Error trying to read the remote lock");
				connectionLost();
			}
			/*
			 * SCP the lock from the remote
			 */
			/*SCPClient scopy = new SCPClient(conn);
			try {
				scopy.get(GeneralUtil.makeSpacesHard(getRemoteFile() 
						+ "/" + LOCK_FILE), getRoot().getAbsolutePath());
			} catch (IOException e) {
				return false;
			}*/
		}
		
		/*
		 * Make sure that there is a file to look at
		 */
		try { 
			if(new File(getRoot(), LOCK_FILE).createNewFile()){
				//unlockMeta();
				return false;
			}
		} catch (IOException e1) { 
			new ErrorDialog(null, "Error trying to create the local lock.\n" +
				"hasMetaLock@Project.java");
		}
		
		/*
		 * read in the lock and see if they match the current
		 */
		String localLock ="";
		try{
			BufferedReader in = new BufferedReader(
					new FileReader(new File(getRoot(), LOCK_FILE)));
			localLock = in.readLine();
			in.close();
		}catch(IOException e){
			new ErrorDialog(null, "Error trying to read local project lock.\n" +
					"hasMetaLock@Project.java");
		}
		
		return localLock != null 
				&& LOCKED_CODE.equals(localLock);
	}
	
	private String getUser(){
		return user;
	}
	
	public Connection getConn(){
		return conn;
	}
	
	public boolean getBrowse(){
		return readonly;
	}
	/**
	 * Get a list of files in the current project directory, local or remote.
	 * @return a vector of strings that hold the full path names of all the files
	 * in the project folder.
	 */
	public Vector <String> getFileList(){
		
		Vector <String> ans = new Vector<String>();
		
		if(isRemote(true) == true){
			try{
			Session sess = conn.openSession();
			sess.execCommand("ls " + GeneralUtil.makeSpacesHard(remotefolder));
			InputStream stdout = new StreamGobbler(sess.getStdout());
			BufferedReader br = new BufferedReader(new InputStreamReader(stdout));
			String line = br.readLine();
			while (line!= null)
			{
				ans.add(getRemoteFile()+"/"+line);
				line = br.readLine();
			}
			sess.close();
		} catch (IOException e) {
			new ErrorDialog(frame, "Error trying to read files from remote folder.\n" +
					"getFileList@Project.java");
		}
		}else{
			File [] temp = getRoot().listFiles();
			for(int i = 0; i < temp.length; i ++){
				ans.add(temp[i].getAbsolutePath());
			}
		}
		return ans;
	
	}
	/**
	 * get the metafile version
	 * @return the version of the current project
	 */
	public String getVersion(){
		return projectInfo.getDocumentElement().getAttribute(VS_KEY);
	}
	
	/**
	 * get the user defined editor command
	 * @return a string identifying the current user
	 * defined editor
	 */
	public String getEditor(){
		return projectInfo.getDocumentElement().getAttribute(USER_EDITOR_KEY);
	}
	
	/**
	 * get the default editor associated with the os type
	 * @return a string contianing the default editor command
	 */
	public String getDefaultEditor(){
		if(System.getProperty("os.name").contains("window")
				||  System.getProperty("os.name").contains("Window")){
			//you use window's you poor poor baby
			return WINDOW_DEFAULT_EDITOR;
		} else if(System.getProperty("os.name").contains("mac")
				|| System.getProperty("os.name").contains("Mac")){
			return MAC_DEFAULT_EDITOR;
		} 	
		else
			return LINUX_DEFAULT_EDITOR;
	}
	
	/**
	 * look at the root
	 * @return the root directory folder
	 */
	public File getRoot(){
		return rootfolder;
	}
	
	/**
	 * get the string that defines the remote project file
	 * @return the path of the remote project directory
	 */
	public String getRemoteFile(){
		return remotefolder;
	}
	
	/**
	 * return the current path of plink
	 * @return returns a string that contains the location of plink
	 */
	public String getPlinkPath(){
		return projectInfo.getDocumentElement().getAttribute(PLINK_PATH_KEY);
	}
	
	/**
	 * Get the full haploview path name
	 * @return a string contianing the location of haploview
	 */
	public String getHaploviewPath(){
		return projectInfo.getDocumentElement().getAttribute(HAPLOVIEW_PATH_KEY);
	}
	
	/**
	 * Get the string that is appended to haploview java launch command
	 * in this project
	 * @return return a string that should be appended to the haploview
	 * launching command
	 */
	public String getHaploviewAppend(){
		return projectInfo.getDocumentElement().getAttribute(HAPLOVIEW_APPEND);
	}
	
	/**
	 * Get any prefix to that belongs before the plink command is passed
	 * @return a string that is the plink prefix
	 */
	public String getPlinkPrefix(){
		return projectInfo.getDocumentElement().getAttribute(PLINK_PREFIX);
	}
	
	/**
	 * get the number of operations in a project.
	 * @return the numbre of operations ever created in this project
	 */
	public int getNumOp(){
		return Integer.parseInt(projectInfo.getDocumentElement().getAttribute(EOP_KEY));
	}
	
	/**
	 * get the name of a project
	 * @return a String which is the name of the project
	 */
	public String getProjectName(){
		if (projectInfo == null)
			return "";
		return projectInfo.getDocumentElement().getAttribute(NAME_KEY);
	}
	
	/**
	 * get the command line for a given operation
	 * @param name the name of the operation you are looking for
	 * @return the command line attached to that operation
	 */
	public String getOpCommand(String name){
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
			if(index.getNodeName().equals(OP_KEY)){
				//check to see if it has the right name
				if(((Element)index).getAttribute(NAME_KEY).equals(name))
					return ((Element)index).getAttribute(CLINE_KEY);
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		
		return null;
	}
	
	/**
	 * get the discription of an operation based on a given name
	 * @param name a string containing the name of the operation
	 * @return a string contianing the discription associated with
	 * the operation
	 */
	public String getOpDiscription(String name){
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type (text, entity or operation)
			if(index.getNodeName().equals(OP_KEY)){
				//check to see if it has the right file/operation name
				if(((Element)index).getAttribute(NAME_KEY).equals(name)){
					return index.getFirstChild().getNodeValue();
				}
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		
		return null;
	}
	
	/**
	 * get a list of all the operation names in the project
	 * @return a vector of all operations names this will be
	 * empty if there are no operations
	 */
	public Vector<String> getAllOpName(){
		//if we don't have any document initalized then logically we shouldn't
		//...return any operations
		if (projectInfo == null)
			return null;
		//create an answer for us to return
		Vector<String> ans = new Vector<String>();
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you fall off
		while(index != null){
			//checks to see if we are at the right type (text, entity or operation)
			if(index.getNodeName().equals(OP_KEY)){
				//grab the name, add it to your answer
				ans.add(((Element) index).getAttribute(NAME_KEY));
			}//end if
			//go to the next sibling
			index = index.getNextSibling();
		}//end while
		//return the ans
		return ans;
	}
	
	/**
	 * returns a list of the files used to construct a given operation
	 * and their corrisponding discriptions.
	 * @param opName the operation name that you are looking for
	 * @param inOut flag that looks at either the INFILE or OUTFILE of
	 * the operation
	 * @return a list of file nodes and text (used form XML formating) nodes
	 */
	public String [] [] getOpFiles(String opName, String inOut){
		String [][] ans = null;
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you fall off
		while(index != null){
			//checks to see if we are at the right type 
			if(index.getNodeName().equals(OP_KEY)){
				//we know that all the operations are elements
				Element temp = (Element) index;
				//check the name
				if(temp.getAttribute(NAME_KEY).equals(opName)){
					//go though the children nodes
					Node index2 = index.getFirstChild();
					while(index2 != null){
						//find the infile/outfile node
						if(index2.getNodeName().equals(inOut)){
							//get the children some of which will be text nodes
							//and some of which will be file nodes
							int numFiles = Integer.parseInt(
									((Element)index2)
									.getAttribute(NUM_FILE_KEY));
							//keep track of the index we are on so that we
							//...can populate the array
							int curIndex = 0;
							//set the answer finally
							ans = new String [2] [numFiles];
							//go through the childre nodes of the file node
							Node index3 = index2.getFirstChild();
							while (index3 != null){
								//make sure that it is a file and 
								//...not a text node
								if(index3.getNodeName().equals(FILE_KEY)){
									ans[0][curIndex] = 
										((Element)index3)
										.getAttribute(NAME_KEY);
									ans[1][curIndex] = index3.getTextContent();
									curIndex ++;
								}
								//get the next possible file node
								index3 = index3.getNextSibling();
							}
							//we went though all the files and can now quit
							return ans;
						}//end if: outfile/infile name check
						
						//go to the next child node of our operation
						index2 = index2.getNextSibling();
					}//end while: going though all second tier children
					
					//we checked all decendents of our matching operation
					return ans;
				}//end if: operation name check
			}//end if: operation type check
			
			//go to the next sibling
			index = index.getNextSibling();
		}//go though all first tier children
		
		return ans;
	}

	
	/**
	 * construct the project by either settting&reading the root directory or
	 * setting all the internal info to null.
	 * @param f diretory that we wish to point the root to
	 */
	public Project(File f, MainFrame mf){
		setBrowse(false);
		frame = mf;
		
		
		
		//if the project is initalized with info
		if (f != null)
			//set the root folder we are working in
			setRoot(f);
		else{
			//set everything to null
			rootfolder = null;
			projectInfo = null;
		}
		
		
	}

	/**
	 * Save, unlock and set the root file to null.
	 *
	 */
	public void closeOldProject(){
		if(projectInfo != null
				&& getRoot() != null){
			writeProjectInfo(true, false);
			
			if(hasMetaLock())
				unlockMeta();
			
			projectInfo = null;
		}	
		
		setBrowse(frame.menuStuff.browseOnly.isSelected());
			
		closeRemote();
		
		setRoot(null);
		
	}
	
	/**
	 * Update all the output files in the project
	 *
	 */
	public void updateFiles(){
		updateFiles(false);
	}
	
	public void updateFiles(boolean forceUpdate){
		
		if(projectInfo == null)
			return;
		
		Vector <String> allLogRoots = new Vector <String>();
		Vector <String> allFiles = getFileList();
		
		for(String files: allFiles){
			if(files.endsWith(".log") 
					|| files.endsWith(OP_STATUS_EXT)){
				String fileName = GeneralUtil.fileName(files);
				if(files.endsWith(".log")){
					allLogRoots.add(fileName.substring(0, fileName.length()-4));
				}
				if( this.isRemote(true) 
						&& ( ! new File(getRoot(), fileName).exists()
						|| forceUpdate == true)){
					SCPClient scopy = new SCPClient(conn);
					try {
						scopy.get(GeneralUtil.makeSpacesHard(files),
								getRoot().getAbsolutePath());
					} catch (IOException e) {
						new ErrorDialog(frame, "Error trying to copy log & gplink files.\n" +
							"Copying ["	+ files + "] to [" 
							+ getRoot().getAbsolutePath() + "] \n" +
							"updateFiles:Project.java");
						
					}
					
				}
			}
		}
		
//		start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
			if(index.getNodeName().equals(OP_KEY)){
				Node index2 = index.getFirstChild();
				
				//go to the Outfiles node
				while(index2 != null
						&& (index2.getNodeName().equals(OUTFILE_KEY) == false )){
					index2 = index2.getNextSibling();
				}
				
				if(index2 == null){
					System.err.println("can't find outfile node " +
							"while updatingFiles");
					return;
				}
					
				
				String opName = ((Element) index).getAttribute(NAME_KEY);
				String [][] outFiles = getOpFiles(opName, OUTFILE_KEY);
				
				String rootName;
				
				
				
				//get a list of all files that have the same root
				rootName = opName;
				
				//File [] allFiles = rootfolder.listFiles();
				
				HashSet<String> filteredFiles = new HashSet <String>();
				
				for(String file: allFiles){
					if((isRemote(false) && file.contains("/" + rootName + "."))
							||(!isRemote(false) && file.contains(File.separator + rootName + ".")))
					{
						filteredFiles.add(file);
					}
				}
				
				Vector<String> temp = new Vector<String>();
				for(String file:filteredFiles){
//				double check that it doesn't match another
					for(String root: allLogRoots){
						if(!opName.equals(root)
								&& root.contains(opName)
								&& new File(file).getName().startsWith(root+"."))
							temp.add(file);
					}
				}
				
				for(String file:temp){
					filteredFiles.remove(file);
				}
				
				//we start with one file
				if(outFiles[0].length >= 1){
					//compare the two lists
					for(int i = 0; i < outFiles[0].length; i ++){
						if(filteredFiles.contains(outFiles[0][i]))
							filteredFiles.remove(outFiles[0][i]);
					}
				} else {
					System.err.println("Problem updating files.");
				}
				//add the new files
				for(String s: filteredFiles){
					addFile(index2, s, "");
				}
				
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		
		return;
	}
	
	/**
	 * load the data from the log file if it exists. If it does not exist
	 * then create a head for the project (note that there is no call to write
	 * to the log file in this operation).
	 *
	 */
	@SuppressWarnings("static-access")
	private void loadData(){
		//double check that the rootfile is not null
		if(rootfolder == null){
			new ErrorDialog(frame, "Error trying to load project data; root folder set to null.\n" +
					"loadData@Project.java");
			//System.err.println("loadData()@Project.java:Project needs to have the root folder set. No data loaded");
			return;
		}
		//create the whole file path of the log file
		File workinglog = new File(rootfolder.getAbsolutePath(), LOG_FILE);
		
		
		//use xml document builders to create a new document from the xml
		//...log file
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setIgnoringComments(true);
		DocumentBuilder parser = null;
		
		Properties localConfig = new Properties();
		
		//only stays this setting if you are in a remote folder
		File configFile = new File(System.getProperty("user.dir")
				+ File.separator + GLOBAL_CONFIG_FILE);
		
		try{
			//System.out.println("loadData@Project.java The config file is: " + configFile.getAbsolutePath());
			localConfig.load(new FileInputStream(configFile));
			
		}catch(Exception e1){
			//System.out.println("loadData@Project.java The config file is: " + configFile.getAbsolutePath());
			configFile = new File(System.getProperty("user.home")
					+ File.separator + GLOBAL_CONFIG_FILE);
			try{
				localConfig.load(new FileInputStream(configFile));
			}catch(Exception e2){
				localConfig.setProperty(PLINK_PATH_KEY, "plink");
				localConfig.setProperty(PLINK_PREFIX, "");
				localConfig.setProperty(HAPLOVIEW_PATH_KEY, "");
				localConfig.setProperty(HAPLOVIEW_APPEND, "");
				localConfig.setProperty(USER_EDITOR_KEY, "");
				localConfig.setProperty(PROJECT1, "");
				localConfig.setProperty(PROJECT2, "");
				localConfig.setProperty(PROJECT3, "");
				localConfig.setProperty(PROJECT4, "");
				localConfig.setProperty(BROWSE_KEY, FALSE);
			}
		}
		
		
		Properties remoteConfig = new Properties();
		if(isRemote(true)){
			try {
				SCPClient scopy = new SCPClient(conn);
				scopy.get("~/" + GLOBAL_CONFIG_FILE, getRoot().getAbsolutePath());
				remoteConfig.load(new FileInputStream(new File(getRoot(), GLOBAL_CONFIG_FILE)));
				
			} catch (IOException e) {
				//it will all get taken care off later we just don't
				//set the remote properties "right" nothing breaks
				if(remoteConfig.getProperty(PLINK_PATH_KEY) == null)
					remoteConfig.setProperty(PLINK_PATH_KEY, "");
				if(remoteConfig.getProperty(PLINK_PREFIX) == null)
					remoteConfig.setProperty(PLINK_PREFIX, "");
			}
			
		}
			
		if(localConfig.getProperty(PROJECT1) == null)
			localConfig.setProperty(PROJECT1, "");
		if(localConfig.getProperty(PROJECT2) == null)
			localConfig.setProperty(PROJECT2, "");
		if(localConfig.getProperty(PROJECT3) == null)
			localConfig.setProperty(PROJECT3, "");
		if(localConfig.getProperty(PROJECT4) == null)
			localConfig.setProperty(PROJECT4, "");
		if(localConfig.getProperty(HAPLOVIEW_PATH_KEY) == null)
			localConfig.setProperty(HAPLOVIEW_PATH_KEY, "");
		if(localConfig.getProperty(HAPLOVIEW_APPEND) == null)
			localConfig.setProperty(HAPLOVIEW_APPEND, "");
		if(localConfig.getProperty(PLINK_PATH_KEY) == null)
			localConfig.setProperty(PLINK_PATH_KEY, "plink");
		if(localConfig.getProperty(PLINK_PREFIX) == null)
			localConfig.setProperty(PLINK_PREFIX, "");
		if(localConfig.getProperty(USER_EDITOR_KEY) == null)
			localConfig.setProperty(USER_EDITOR_KEY, "");
		if(localConfig.getProperty(BROWSE_KEY) == null)
			localConfig.setProperty(BROWSE_KEY, FALSE);
		
		//System.out.println(config.get(PROJECT1));
		if(!localConfig.get(PROJECT1).equals(rootfolder.getAbsolutePath())){
			if(localConfig.get(PROJECT2).equals(getRoot().getAbsolutePath())){
				localConfig.setProperty(PROJECT2, localConfig.getProperty(PROJECT1));
			}else if(localConfig.get(PROJECT3).equals(getRoot().getAbsolutePath())){
				localConfig.setProperty(PROJECT3, localConfig.getProperty(PROJECT2));
				localConfig.setProperty(PROJECT2, localConfig.getProperty(PROJECT1));
			} else{
				localConfig.setProperty(PROJECT4, localConfig.getProperty(PROJECT3));
				localConfig.setProperty(PROJECT3, localConfig.getProperty(PROJECT2));
				localConfig.setProperty(PROJECT2, localConfig.getProperty(PROJECT1));
			}
				
			localConfig.setProperty(PROJECT1, getRoot().getAbsolutePath());
			
			try{
				//System.out.println(configFile.getAbsolutePath());
				localConfig.store(new FileOutputStream(configFile), 
						"gPLINK global configuration");
			}catch(IOException e1){
				new ErrorDialog(frame, "Error storing configuration file.\n" +
						"loadData@Project.java");
				//return;
			}
			
			if(frame != null 
					&& frame.menuStuff != null)
				frame.menuStuff.setPreviousProjects(localConfig);
		}
		
		
		//copy the log file from the remote folder
		if(isRemote(true)){
			if(workinglog.exists() ){
				//delete any local copy of the metafile we have
				workinglog.delete();
			}
			
			try {
				//copy the metafile
				SCPClient scopy = new SCPClient(conn);
				scopy.get(GeneralUtil.makeSpacesHard(remotefolder 
						+ "/" + LOG_FILE), getRoot().getAbsolutePath());
			} catch (IOException e) {
				new ErrorDialog(frame, "Error trying to copy the log file for this" +
						" remote project.\n loadData@Project.java");
				//if the file isn't there then move on
				
			}
		}
		
		try {
			parser = factory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			new ErrorDialog(frame, "Error trying to create a new document builder.\n" +
					"loadData@Project.java");
		}
		
		//take the remote parames that are appropreate
		Properties config = localConfig;
		if(isRemote(true)){
			config.setProperty(PLINK_PATH_KEY, remoteConfig.getProperty(PLINK_PATH_KEY));
			config.setProperty(PLINK_PREFIX, remoteConfig.getProperty(PLINK_PREFIX));
		}
		
		
		//check that the log file exists
		if(workinglog.exists() ){
			try {
				projectInfo = parser.parse(workinglog);
			} catch (SAXException e) {
				new ErrorDialog(frame, "Error trying to create project, " +
					"can not parse log file.\n Creating new " +
					"project, please import old operations.\n " +
					"loadData@Project.java");
				createNewProject(parser);
			} catch (IOException e) {
				new ErrorDialog(frame, "Error trying to create project, " +
					"IOException.\n Creating new" +
					"project, please import old operations.\n " +
					"loadData@Project.java");
				createNewProject(parser);
			}
			projectInfo.getDocumentElement().setAttribute(VS_KEY, frame.VS);
		}else{
			createNewProject(parser);
		}
		
		setPlinkPath(config.getProperty(PLINK_PATH_KEY));
		setPlinkPrefix(config.getProperty(PLINK_PREFIX));
		setHaploviewPath(config.getProperty(HAPLOVIEW_PATH_KEY));
		setHaploviewAppend(config.getProperty(HAPLOVIEW_APPEND));
		setEditor(config.getProperty(USER_EDITOR_KEY));
		setBrowse(config.getProperty(BROWSE_KEY).equals(TRUE));
		
		if(!projectInfo.getDocumentElement().getAttribute(HOME_KEY).equals(
				rootfolder.getAbsolutePath())
				&& isRemote(false) == false)
			replaceHome();
		
		
	}
	
	/**
	 * Create a new project from scratch.
	 * @param parser is a document builder
	 * configuration information for this project
	 */
	@SuppressWarnings("static-access")
	private void createNewProject(DocumentBuilder parser){

		//create a new document
		projectInfo = parser.newDocument();
		
		//put in the header information for the project but no file data
		projectInfo.createTextNode(String.format("%n", (Object []) null));
		Element root = 
			(Element) projectInfo.createElement("project");
		root.setAttribute(NAME_KEY, rootfolder.getName());
		root.setAttribute(VS_KEY, frame.VS);
		root.setAttribute(EOP_KEY, "0");
		root.setAttribute(HOME_KEY, rootfolder.getAbsolutePath());
		
		//apppend the root to the project
		projectInfo.appendChild(root);
		//pad the file with new lines to make it human readable
	    root.appendChild( projectInfo.createTextNode(String.format("%n", (Object []) null)));
	
	}
	
	/**
	 * Check with the user to see if they want the old home file instances
	 * replaces with the new one.
	 *
	 */
	public void replaceHome(){
		int choice = JOptionPane.showConfirmDialog(null,
				"I detect a different path. " +
				"Do you want to replace all instances of: \n[" +
				projectInfo.getDocumentElement().getAttribute(HOME_KEY) +
				"] with [" + rootfolder.getAbsolutePath()+ "]?\n" +
						"(Hint: if you have moved your folder select yes.)", 
				"Replace?", JOptionPane.YES_NO_OPTION);
		if(choice == JOptionPane.NO_OPTION)
			return;
		if(choice == JOptionPane.YES_OPTION){
			
			for(String opName: getAllOpName()){
				String command = getOpCommand(opName);
				String discr = getOpDiscription(opName);
				String oldHome  = projectInfo.getDocumentElement().getAttribute(HOME_KEY);
	        	String [][] infiles = getOpFiles(opName, INFILE_KEY);
	        	for(int i = 0; i < infiles[0].length; i ++){
	        		String root = new File(infiles[0][i]).getParent();
	        		if(root.equals(oldHome))
	        		infiles[0][i] = new File(rootfolder, 
	        				new File(infiles[0][i]).getName()).getAbsolutePath();
	        	}
	        	
	        	String [][] outfiles = getOpFiles(opName, OUTFILE_KEY);
        		for(int i = 0; i < outfiles[0].length; i ++){
        			String root = new File(outfiles[0][i]).getParent();
	        		if(root.equals(oldHome))
        			outfiles[0][i] = new File(rootfolder, 
	        				new File(outfiles[0][i]).getName()).getAbsolutePath();
	        	}
        		
        		removeOperation(opName);
        		addOperation(command, opName, discr, 
						infiles[0], infiles[1],
						outfiles[0], outfiles[1]);
	        }
			
			projectInfo.getDocumentElement().setAttribute(HOME_KEY, 
					rootfolder.getAbsolutePath());
			
			writeProjectInfo(true, false);
			
		}
			
	}
	
	private Node getOperation(String name){
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
			if(index.getNodeName().equals(OP_KEY)){
				//check to see if it has the right name
				if(((Element)index).getAttribute(NAME_KEY).equals(name))
					return index;
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		return null;
	}
	
	public void swapOp(String op1, String op2){
		
		Node firstNode = getOperation(op1);
		Node secondNode = getOperation(op2);
		
		//replace the seond node with the first and store the parentless old node
		secondNode = projectInfo.getDocumentElement()
						.replaceChild(firstNode.cloneNode(true), secondNode);
		
		projectInfo.getDocumentElement()
					.replaceChild(secondNode.cloneNode(true), firstNode);
		
	}
	
	/**
	 * add a file element to a given parent parent node with their discription.
	 * @param parent node that we are attaching the file-node too
	 * @param name set the name attribute of this file-node
	 * @param discription set the text of the file node
	 */
	void addFile(Node parent, String name, String discription){
		
		if(parent == null
				|| parent.getNodeType() != (Node.ELEMENT_NODE)){
			System.err.println("We are trying to add a file to a " +
					"node which can not accept it.");
			return;
		}
		
		//new file node
		Element newFile = (Element) projectInfo.createElement(FILE_KEY);
		//set the name attribute
		newFile.setAttribute(NAME_KEY, name);
		//set the text content
		//newFile.setTextContent(discription);
		newFile.appendChild(projectInfo.createTextNode(discription));
		//set up spacing
		parent.appendChild( projectInfo.createTextNode(String.format("%n",(Object[]) null)));
		//attach the file node
		parent.appendChild(newFile);
		//grab the file count in the parent
		int num_files = Integer.parseInt(
				 ((Element)parent).getAttribute(NUM_FILE_KEY));
		//add this to the file count in the parent
		((Element)parent).setAttribute(NUM_FILE_KEY, "" + (num_files +1));
				
	}
	
	/**
	 * replace the discription of a file with a new discription
	 * @param parent parent node of the file we are looking for
	 * @param name name of the file we are looking for
	 * @param discription new discription of file
	 */
	void replaceFile(Node parent, String name, String discription){
		//dummy check the inputs
		if(parent == null
				|| parent.getNodeType() != (Node.ELEMENT_NODE)){
			System.err.println("We are trying to add a file to a " +
					"node which can not accept it.");
			return;
		}
		
		//go though the children of the parent node
		Node index = parent.getFirstChild();
		while(index != null){
			//check that we are at a file node with the right name
			if(index.getNodeName().equals(FILE_KEY)
					&& ((Element) index).getAttribute(NAME_KEY).equals(name)){
				//there might not be any children
				if(index.getFirstChild() != null)
					//reset the discription
					index.getFirstChild().setNodeValue(discription);
				else
					//add a new child with the discription
					index.appendChild(projectInfo.createTextNode(discription));
				return;
			}
			index = index.getNextSibling();
		}
		
		//System.err.println("Could not find the file to replace " +
		//		"the discription for.");
		return;
	}
	
	/**
	 * add an operation element to the project element.
	 * @param cline the actual command line that is sent plink
	 * @param name name that command for internal reference
	 * @param discription set the discription of that command
	 * @param input an array of names of the input files
	 * @param inputDis an array of discriptions corrisponding to the inptu files
	 * @param output an array of names of the output files
	 * @param outputDis an array of discriptions corrisponding to the output files
	 * @return return that operation element that you created incase 
	 * you want to do something with it
	 */
	public Element addOperation(String cline, String name, String discription,
						String [] input, String [] inputDis,
						String [] output, String [] outputDis){
		//create new element to hold operation
		Element newOp = (Element) projectInfo.createElement(OP_KEY);
		//set the appropreate attributes
		newOp.setAttribute(CLINE_KEY, cline);
		newOp.setAttribute(NAME_KEY, name);
		if(gplinkLog != null)
			gplinkLog.addOp(name);
		//set the text for the node as the discription
		newOp.setTextContent(discription);
		//append the operation to the project
		projectInfo.getDocumentElement().appendChild(newOp);
		
		//create a new element to hold the infiles
		Element infiles = (Element) projectInfo.createElement(INFILE_KEY);
		//start with no file's
		infiles.setAttribute(NUM_FILE_KEY, "0");
		//set a name key for the hell of it, don't know that it's actually
		//...used
		infiles.setAttribute(NAME_KEY, "Input Files");
		//go through and add each input file name and discription
		for(int i = 0; i < input.length; i ++){
			addFile(infiles, input[i], inputDis[i]);
		}
		//format so that humans can read the xml format
		newOp.appendChild( projectInfo.createTextNode(String.format("%n", (Object[]) null)));
		//add all the infiles to the new operation
		newOp.appendChild(infiles);
		
		//create the new element to hold the outfiles
		Element outfiles = (Element) projectInfo.createElement(OUTFILE_KEY);
		outfiles.setAttribute(NUM_FILE_KEY, "0");
		outfiles.setAttribute(NAME_KEY, "Output Files");
		for(int i = 0; i < output.length; i ++){
			addFile(outfiles, output[i], outputDis[i]);
		}
		
		//format so that humans can read the xml format
		newOp.appendChild( projectInfo.createTextNode(String.format("%n",(Object[]) null)));
		newOp.appendChild(outfiles);
		//format so that humans can read the xml format
		newOp.appendChild( projectInfo.createTextNode(String.format("%n",(Object[]) null)));
		
		//get the number of operations ever created
		int num_ever_op = Integer.parseInt(
			 projectInfo.getDocumentElement().getAttribute(EOP_KEY)) + 1;
		//add this to the operation count
		projectInfo.getDocumentElement().setAttribute(EOP_KEY, "" + num_ever_op);
		
		//return the element we have just created
		return newOp;
	}
	
	/**
	 * replace an old operation with a new one
	 * @param oldname name of the old operation
	 * @param cline new command line you want to add
	 * @param name new name of operation
	 * @param discription new discription
	 * @param input new input file names
	 * @param inputDis new input file discriptions
	 * @param output new output file names
	 * @param outputDis new output file discriptions
	 * @return return a pointer to the node that was modified
	 */
	public Node replaceOperation(String oldname, String cline, String name, 
						String discription, 
						String [] input, String [] inputDis,
						String [] output, String [] outputDis){

		//******Find the operation based on it's old name********
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//element that we are looking for
		Element renewOp = null;
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type and have the right name
			if(index.getNodeName().equals(OP_KEY)
				&& ((Element)index).getAttribute(NAME_KEY).equals(oldname)){
				renewOp = (Element)index;
				}
			//go to the next sibling
			if(index != null)
				index = index.getNextSibling();
		}
		
		if(renewOp == null){
			System.err.println("Can't find the operation to replace.");
			return null;
		}
		
		//********Reset the data***********
		//set the appropreate attributes
		renewOp.setAttribute(CLINE_KEY, cline);
		renewOp.setAttribute(NAME_KEY, name);
		
		//set the text for the node as the discription
		renewOp.getFirstChild().setNodeValue(discription);
		
		index = renewOp.getFirstChild();
		
		while(index != null){
			if(index.getNodeName().equals(INFILE_KEY)){
				for(int i = 0; i < input.length; i ++){
					replaceFile(index, input[i], inputDis[i]);
				}
			}
			if(index.getNodeName().equals(OUTFILE_KEY)){
				for(int i = 0; i < output.length; i ++){
					replaceFile(index, output[i], outputDis[i]);
				}
			}
			index = index.getNextSibling();
		}
		
		//return the element we have just replaced
		return renewOp;
		
	}
	
	/**
	 * delete an operation in the document by name
	 * @param name of the operation to be delete
	 * @return the operation that we just deleted incase we want it
	 */
	public Node removeOperation(String name){
		//start at the first child
		Node index = projectInfo.getDocumentElement().getFirstChild();
		//go until you are at the last child
		while(index != null){
			//checks to see if we are at the right type
			//...and that the name matches
			if(index.getNodeName().equals(OP_KEY)
					&& ((Element) index).getAttribute(NAME_KEY).equals(name)){
				//found it!
				return projectInfo.getDocumentElement().removeChild(index);
			}
			//go to the next sibling
			index = index.getNextSibling();
		}
		//we didn't find the operation
		return null;
	}
	
	/**
	 * This writes the data that makes up the project 
	 * (files and operations in general)
	 * to either the log file (.metafile) and/or the output screen.
	 * @param fileFlag flags the print to file output
	 * @param screenFlag flags the print to screen output
	 * 
	 */
	public void writeProjectInfo(boolean fileFlag, boolean screenFlag){
		
		if(avalibleLock() == false
				&& hasMetaLock() == false 
				&& getBrowse() == false){
			int choice = JOptionPane.showConfirmDialog(frame, 
					"Some other instance of gPLINK has the lock. " +
					"Would you like to steal the " +
					"lock to save the project information?\n If you don't then you will be unable to add any new operations. ", 
					"Steal?", JOptionPane.YES_NO_OPTION);
			if(choice == JOptionPane.CLOSED_OPTION
					|| choice == JOptionPane.NO_OPTION){
				setBrowse(true);
				frame.menuStuff.setMenuLock(!getBrowse());
			}
		}
		
		//System.out.println("writeProjectInfo@Project.java:lock check completed... moving on.");
		
		if(getBrowse() == true){
			return;
		}
		lockMeta();
		frame.menuStuff.setMenuLock(true);
		
//		 Use a Transformer for output
		   TransformerFactory tFactory = TransformerFactory.newInstance();
		   Transformer transformer = null;
	
			//configure the transformer
			try {
				transformer = tFactory.newTransformer();
			} catch (TransformerConfigurationException e1) {
				new ErrorDialog(frame, "Error trying to create transformer.\n" +
						"writeProjectInfo@Project.java");
			}
			//set the source of the data
			DOMSource source = new DOMSource(projectInfo);
			//check the flags to see where we should output
			if(screenFlag == true){
				//set the recieving stream to the output
				StreamResult result = new StreamResult(System.out);
				try {
					transformer.transform(source, result);
				} catch (TransformerException e) {
					new ErrorDialog(frame, "Error trying to transform source to" +
							"console.\n writeProjectInfo@Project.java");
				}
			}
			if(fileFlag == true){
				//set the recieving stream to the file specified by 
				//...rootfolder/logfile
				StreamResult ofile = null;
				try {
					ofile = new StreamResult(new FileWriter(
								new File(rootfolder, LOG_FILE)) );
				} catch (IOException e) {
					new ErrorDialog(frame, "Error trying to create file stream.\n" +
						"writeProjectInfo@Project.java");
				}
				try {
					transformer.transform(source, ofile);
				} catch (TransformerException e) {
					new ErrorDialog(frame, "Error trying to transform source to" +
					"file stream.\n writeProjectInfo@Project.java");
				}
			}
			
		if(isRemote(true)){
			SCPClient scopy = new SCPClient(conn);
			try {
				scopy.put(getRoot().getAbsolutePath() + File.separator + LOG_FILE,
						GeneralUtil.makeSpacesHard(remotefolder), SCP_MODE);
			} catch (IOException e) {
				new ErrorDialog(frame, "Error trying to copy local log file to remote" +
						" project.\n writeProjectInfo@Project.java");
			}
		}
		   
	}
	
	public void writeOperations(boolean fileFlag, boolean screenFlag){
		if(fileFlag == true){

			DateFormat dateFormat = new SimpleDateFormat("yyyy-MMM-dd-HH-mm-ss");
            java.util.Date date = new java.util.Date();
            File outputFile = new File(getRoot(), OP_SUM_FILE+dateFormat.format(date)+".html");
		
			try {
		        BufferedWriter out = new BufferedWriter(new FileWriter(outputFile));
		        out.write("<html><title>" + getProjectName()+"</title><body>");
		        out.write("<h2>"+getProjectName()+"</h2>");
		        out.write("<h3>Timestamp: " + dateFormat.format(date)+ "</h3>");
		        out.write("<h3>local project file location: " + getRoot() + "</h3>");
		        if(isRemote(false))
		        	out.write("<h3>remote project file location: " + getRemoteFile() + "</h3>");
		        out.newLine();
		        out.write("<ul>");
		        
		        for(String opName: getAllOpName()){
		        	out.write("<li>");
		        	out.write(opName+ ": " + getOpDiscription(opName));
		        	
		        	out.newLine();
		        	out.write("<ul>");
		        	out.write("<li>"+getOpCommand(opName)+"</li>");
		        	out.newLine();
		        			        	
		        	String [][] infiles = getOpFiles(opName, INFILE_KEY);
		        	out.write("<li>INFILES<ul>");
		        	out.newLine();
		        	for(int i = 0; i < infiles[0].length; i ++){
		        		out.write("<li>"+infiles[0][i] + ": " + infiles[1][i]+"</li>");
		        		out.newLine();
		        	}
		        	out.write("</ul></li>");
		        	
		        	String [][] outfiles = getOpFiles(opName, OUTFILE_KEY);
		        	out.write("<li>OUTFILES<ul>");
		        	out.newLine();
	        		for(int i = 0; i < outfiles[0].length; i ++){
	        			if(!outfiles[0][i].endsWith(OP_STATUS_EXT)){
			        		out.write("<li>"+outfiles[0][i] + ": " + outfiles[1][i]+"</li>");
			        		out.newLine();
	        			}
		        	}
	        		out.write("</ul></li>");
	        		
	        		out.write("</ul>");
	        		out.write("</li>");
	        		out.newLine();
		        }
		        out.write("</ul>");
        		out.newLine();
        		out.write("</body></html>");
		        out.close();
		    } catch (IOException e) {
		    	new ErrorDialog(frame, "Error trying to write the operations \n" +
		    			"summary.\n writeOperations@Project.java");
		    }
		}
		if(screenFlag == true){
			System.out.println(getProjectName());
			System.out.println();
			for(String opName: getAllOpName()){
	        	System.out.print(opName+ ": " + getOpDiscription(opName));
	        	
	        	String [][] infiles = getOpFiles(opName, INFILE_KEY);
	        	System.out.println("INFILES");
	        	for(int i = 0; i < infiles[0].length; i ++){
	        		System.out.println(infiles[0][i] + ": " + infiles[1][i]);
	        	}
	        	
	        	String [][] outfiles = getOpFiles(opName, OUTFILE_KEY);
	        	System.out.println("OUTFILES");
        		for(int i = 0; i < outfiles[0].length; i ++){
        			System.out.println(outfiles[0][i] + ": " + outfiles[1][i]);
	        	}
        		System.out.println();
	        }
		}		
	}

}
